/* Copyright (C) 2009 Niels Mller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include <stdio.h>

#include <alloca.h>

#include "huffman.h"

/* Gives the number of code words of a given bit length, the first
   code word, and the index of that code word. Code words are selected
   and ordered according to RFC 1951 (lexicographic order, shorter
   codes precede longer, codes of the same length are consecutive). */
struct code_info {
  unsigned short count;
  unsigned short index;
  unsigned short code;
};

/* Locates the largest code word, of length at most bits, that matches
   a given prefix. */
static void
huffman_largest_code (const struct code_info *info,
		      unsigned bits,
		      unsigned prefix_length, unsigned prefix,
		      unsigned *code_length, unsigned *code)
{
  assert(prefix_length > 0);

  /* Find the largest code word matching the prefix. */
  for (; bits >= prefix_length; bits--)
    if (info[bits].count)
      {
	/* The largest word of length bits that matches the prefix. */
	unsigned prefix_max = ((prefix+1) << (bits - prefix_length)) - 1;

	if (info[bits].code <= prefix_max)
	  {
	    unsigned index;

	    /* Index relative to the first code of this length. */
	    index = prefix_max - info[bits].code;
	    if (index >= info[bits].count)
	      index = info[bits].count - 1;

	    *code_length = bits;
	    *code = info[bits].code + index;

	    return;
	  }
      }
  abort();
}

/* Construct a huffman tree representing the codes starting with a
   given prefix. BITS and LAST_CODE are the last code word that
   matches the prefix. Returns the number of used nodes. */
static unsigned
huffman_build_tree(struct huffman_node *tree,
		   const struct code_info *info,
		   /* VALUE[k] gives the value for the k'th code word. */
		   const unsigned short *value,
		   unsigned prefix_length,
		   unsigned bits, unsigned largest_code)
{
  assert (prefix_length <= bits);

  /* We have the largest code word matching the prefix. If the next
     bit of this code word is zero, that means that there's no choice,
     the zero is mandatory. */
  tree->zeros = 0;

  while (prefix_length < bits
	 && !((largest_code >> (bits - prefix_length - 1)) & 1))
    {
      tree->zeros++;
      prefix_length++;
    }

  if (bits == prefix_length)
    {
      /* A leaf node */
      tree->leaf = 1;
      tree->value = value[info[bits].index + largest_code - info[bits].code];
      return 1;
    }
  else
    {
      unsigned prefix;
      unsigned bits_left;
      unsigned largest_left;
      unsigned left;
      unsigned right;

      prefix = largest_code >> (bits - prefix_length);
      
      huffman_largest_code (info, bits, prefix_length + 1, prefix << 1,
			    &bits_left, &largest_left);
      
      left = huffman_build_tree(tree + 1,
				info, value,
				prefix_length + 1,
				bits_left, largest_left);
      
      right = huffman_build_tree(tree + 1 + left,
				 info, value,
				 prefix_length + 1,
				 bits, largest_code);

      tree->leaf = 0;
      tree->value = 1 + left;
      return 1 + left + right;
    }
}

#define MAX_LENGTH 15

int
huffman_tree(struct huffman_node *tree,
	     unsigned n, const unsigned char *length)
{
  struct code_info info[MAX_LENGTH + 1];
  unsigned short index[MAX_LENGTH + 1];
  unsigned bits;
  unsigned k;
  
  /* Gives the value for the j'th code word, with code words ordered
     lexicographically. */
  unsigned short *value;
  unsigned size;

  assert (n > 0);
  memset(info, 0, sizeof(info));

  for (k = 0; k < n; k++)
    {
      if (length[k] > MAX_LENGTH)
	return 0;
      info[length[k]].count++;
    }

  for (bits = MAX_LENGTH; !info[bits].count; bits--)
    ;

  if (!bits)
    /* No code words? All specified to have length zero. */
    return 0;

  /* NOTE: The algorithm in RFC 1951 seem to have an off-by-one error.
     We should not allocate any codes of length zero. */
  info[1].code = 0;
  info[1].index = 0;

  for (k = 2; k <= bits; k++)
    {
      info[k].code = (info[k-1].code + info[k-1].count) << 1;
      info[k].index = info[k-1].index + info[k-1].count;

      /* Make sure code fits in given size. */
      if (info[k].code + info[k].count > (1U << k))
	return 0;
    }
  
  value = alloca((n - info[0].count)  * sizeof(*value));
  memset(index, 0, sizeof(index));

  for (k = 0; k < n; k++)
    if (length[k] > 0)
      value[info[length[k]].index + index[length[k]]++] = k;
    
  size = huffman_build_tree(tree, info, value, 0,
			    bits, info[bits].code + info[bits].count - 1);
  assert(size == HUFFMAN_SIZE(n - info[0].count));

  /* Sanity check */
  for (k = 0; k < size; k++)
    assert(tree[k].leaf || k + tree[k].value < size);

  return 1;
}

void
huffman_init (struct huffman_ctx *ctx,
	      const struct huffman_node *root)
{
  assert(!root->leaf ||root->zeros);

  ctx->root = root;
  ctx->p = root;
  ctx->zeros = root->zeros;
}

int
huffman_decode_bits (struct huffman_ctx *ctx,
		     struct bit_buffer *input,
		     unsigned *value)
{
  assert(input->n > 0);

  do
    {
      unsigned bit = input->bits & 1;
      input->bits >>= 1;
      input->n--;

      if (ctx->zeros > 0)
	{
	  /* Next bit must be zero. */
	  if (bit)
	    return -1;
      
	  ctx->zeros--;
	}
      else
	{
	  assert(!ctx->p->leaf);

	  ctx->p += bit ? ctx->p->value : 1;
	  ctx->zeros = ctx->p->zeros;
	}
      
      if (ctx->zeros == 0 && ctx->p->leaf)
	{
	  *value = ctx->p->value;

	  ctx->p = ctx->root;
	  ctx->zeros = ctx->root->zeros;

	  return 1;
	}
    }
  while (input->n > 0);
  
  return 0;
}
